@@ -14,14 +14,12 @@
 import logging
 from functools import partial
 
-from botocore import xform_name
-
 from .action import ServiceAction
 from .action import WaiterAction
-from .base import ServiceResource
+from .base import ResourceMeta, ServiceResource
 from .collection import CollectionFactory
 from .model import ResourceModel
-from .response import all_not_none, build_identifiers
+from .response import build_identifiers, ResourceHandler
 from ..exceptions import ResourceLoadException
 
 
@@ -36,8 +34,9 @@ class ResourceFactory(object):
     SQS resource) and another on models contained within the service (e.g. an
     SQS Queue resource).
     """
-    def __init__(self):
+    def __init__(self, emitter):
         self._collection_factory = CollectionFactory()
+        self._emitter = emitter
 
     def load_from_definition(self, service_name, resource_name, model,
                              resource_defs, service_model):
@@ -64,30 +63,32 @@ class ResourceFactory(object):
         :rtype: Subclass of :py:class:`~boto3.resources.base.ServiceResource`
         :return: The service or resource class.
         """
-        # Set some basic info
-        meta = {
-            'service_name': service_name,
-            'identifiers': [],
-            'data': None,
-        }
-        attrs = {
-            'meta': meta,
-        }
+
 
         logger.debug('Loading %s:%s', service_name, resource_name)
 
         resource_model = ResourceModel(resource_name, model, resource_defs)
 
+        shape = None
+        if resource_model.shape:
+            shape = service_model.shape_for(resource_model.shape)
+        resource_model.load_rename_map(shape)
+
+
+        # Set some basic info
+        meta = ResourceMeta(service_name, resource_model=resource_model)
+        attrs = {
+            'meta': meta,
+        }
+
         self._load_identifiers(attrs, meta, resource_model)
-        self._load_subresources(attrs, service_name, resource_name,
-                                resource_model, resource_defs, service_model)
         self._load_actions(attrs, resource_model, resource_defs,
                            service_model)
         self._load_attributes(attrs, meta, resource_model, service_model)
         self._load_collections(attrs, resource_model, resource_defs,
                                service_model)
-        self._load_references(attrs, service_name, resource_name,
-                              resource_model, resource_defs, service_model)
+        self._load_has_relations(attrs, service_name, resource_name,
+                                 resource_model, resource_defs, service_model)
         self._load_waiters(attrs, resource_model)
 
         # Create the name based on the requested service and resource
@@ -95,7 +96,12 @@ class ResourceFactory(object):
         if service_name != resource_name:
             cls_name = service_name + '.' + cls_name
 
-        return type(str(cls_name), (ServiceResource,), attrs)
+        base_classes = [ServiceResource]
+        if self._emitter is not None:
+            self._emitter.emit('creating-resource-class.%s' % cls_name,
+                               class_attributes=attrs,
+                               base_classes=base_classes)
+        return type(str(cls_name), tuple(base_classes), attrs)
 
     def _load_identifiers(self, attrs, meta, model):
         """
@@ -104,38 +110,8 @@ class ResourceFactory(object):
         operations on the resource.
         """
         for identifier in model.identifiers:
-            snake_cased = xform_name(identifier.name)
-            snake_cased = self._check_allowed_name(
-                attrs, snake_cased, 'identifier', model.name)
-            meta['identifiers'].append(snake_cased)
-            attrs[snake_cased] = None
-
-    def _load_subresources(self, attrs, service_name, resource_name,
-                           model, resource_defs, service_model):
-        """
-        Creates subresource classes which hang off the instance. Each
-        subresource is a bound partial method that returns a resource
-        instance which shares the client and identifiers of the parent.
-        """
-        # Create dangling classes, e.g. SQS.Queue, SQS.Message
-        if service_name == resource_name:
-            # This is a service, so dangle all the resource_defs as if
-            # they were subresources of the service itself.
-            for name, resource_def in resource_defs.items():
-                cls = self.load_from_definition(
-                    service_name, name, resource_defs.get(name, {}),
-                    resource_defs, service_model)
-                attrs[name] = self._create_class_partial(cls)
-
-        # For non-services, subresources are explicitly listed
-        if model.sub_resources:
-            identifiers = model.sub_resources.identifiers
-            for name in model.sub_resources.resource_names:
-                cls = self.load_from_definition(
-                    service_name, name, resource_defs.get(name, {}),
-                    resource_defs, service_model)
-                attrs[name] = self._create_class_partial(
-                    cls, identifiers=identifiers)
+            meta.identifiers.append(identifier.name)
+            attrs[identifier.name] = None
 
     def _load_actions(self, attrs, model, resource_defs, service_model):
         """
@@ -145,16 +121,12 @@ class ResourceFactory(object):
         """
         if model.load:
             attrs['load'] = self._create_action(
-                'load', model.load, resource_defs, service_model,
-                is_load=True)
+                model.load, resource_defs, service_model, is_load=True)
             attrs['reload'] = attrs['load']
 
         for action in model.actions:
-            snake_cased = xform_name(action.name)
-            snake_cased = self._check_allowed_name(
-                attrs, snake_cased, 'action', model.name)
-            attrs[snake_cased] = self._create_action(snake_cased,
-                action, resource_defs, service_model)
+            attrs[action.name] = self._create_action(action, resource_defs,
+                                                     service_model)
 
     def _load_attributes(self, attrs, meta, model, service_model):
         """
@@ -166,16 +138,9 @@ class ResourceFactory(object):
         if model.shape:
             shape = service_model.shape_for(model.shape)
 
-            for name, member in shape.members.items():
-                snake_cased = xform_name(name)
-                if snake_cased in meta['identifiers']:
-                    # Skip identifiers, these are set through other means
-                    continue
-
-                snake_cased = self._check_allowed_name(
-                    attrs, snake_cased, 'attribute', model.name)
-                attrs[snake_cased] = self._create_autoload_property(name,
-                    snake_cased)
+            attributes = model.get_attributes(shape)
+            for name, (orig_name, member) in attributes.items():
+                attrs[name] = self._create_autoload_property(orig_name, name)
 
     def _load_collections(self, attrs, model, resource_defs, service_model):
         """
@@ -185,36 +150,39 @@ class ResourceFactory(object):
         through the collection's items.
         """
         for collection_model in model.collections:
-            snake_cased = xform_name(collection_model.name)
-            snake_cased = self._check_allowed_name(
-                attrs, snake_cased, 'collection', model.name)
-
-            attrs[snake_cased] = self._create_collection(
-                attrs['meta']['service_name'], model.name, snake_cased,
+            attrs[collection_model.name] = self._create_collection(
+                attrs['meta'].service_name, model.name,
                 collection_model, resource_defs, service_model)
 
-    def _load_references(self, attrs, service_name, resource_name,
-                         model, resource_defs, service_model):
+    def _load_has_relations(self, attrs, service_name, resource_name,
+                            model, resource_defs, service_model):
         """
-        Load references, which are related resource instances. For example,
-        an EC2 instance would have a ``vpc`` reference, which is an instance
-        of an EC2 VPC resource.
+        Load related resources, which are defined via a ``has``
+        relationship but conceptually come in two forms:
+
+        1. A reference, which is a related resource instance and can be
+           ``None``, such as an EC2 instance's ``vpc``.
+        2. A subresource, which is a resource constructor that will always
+           return a resource instance which shares identifiers/data with
+           this resource, such as ``s3.Bucket('name').Object('key')``.
         """
         for reference in model.references:
-            snake_cased = xform_name(reference.resource.type)
-            snake_cased = self._check_allowed_name(
-                attrs, snake_cased, 'reference', model.name)
-            attrs[snake_cased] = self._create_reference(
-                reference.resource.type, snake_cased, reference, service_name,
-                resource_name, model, resource_defs, service_model)
+            # This is a dangling reference, i.e. we have all
+            # the data we need to create the resource, so
+            # this instance becomes an attribute on the class.
+            attrs[reference.name] = self._create_reference(
+                reference.resource.type, reference,
+                service_name, resource_name, model, resource_defs,
+                service_model)
+
+        for subresource in model.subresources:
+            # This is a sub-resource class you can create
+            # by passing in an identifier, e.g. s3.Bucket(name).
+            name = subresource.resource.type
+            attrs[subresource.name] = self._create_class_partial(
+                name, subresource, service_name, resource_name, model,
+                resource_defs, service_model)
 
-        for reference in model.reverse_references:
-            snake_cased = xform_name(reference.resource.type)
-            snake_cased = self._check_allowed_name(
-                attrs, snake_cased, 'reference', model.name)
-            attrs[snake_cased] = self._create_reference(
-                reference.resource.type, snake_cased, reference, service_name,
-                resource_name, model, resource_defs, service_model)
 
     def _load_waiters(self, attrs, model):
         """
@@ -223,44 +191,7 @@ class ResourceFactory(object):
         of the resource.
         """
         for waiter in model.waiters:
-            snake_cased = xform_name(waiter.resource_waiter_name)
-            snake_cased = self._check_allowed_name(
-                attrs, snake_cased, 'waiter', model.name)
-            attrs[snake_cased] = self._create_waiter(waiter, snake_cased)
-
-    def _check_allowed_name(self, attrs, name, category, resource_name):
-        """
-        Determine if a given name is allowed on the instance, and if not,
-        then raise an exception. This prevents public attributes of the
-        class from being clobbered, e.g. since we define ``Resource.meta``,
-        no identifier may be named ``meta``. Another example: no action
-        named ``queue_items`` may be added after an identifier of the same
-        name has been added.
-
-        One attempt is made in the event of a collision to remedy the
-        situation. The ``category`` is appended to the name and the
-        check is performed again. For example, if an action named
-        ``get_frobs`` fails the test, then we try ``get_frobs_action``
-        after logging a warning.
-
-        :raises: ValueError
-        """
-        if name in attrs:
-            logger.warning('%s `%s` would clobber existing %s'
-                           ' resource attribute, going to try'
-                           ' %s instead...', category, name,
-                           resource_name, name + '_' + category)
-            # TODO: Move this logic into the model and strictly
-            #       define the loading order of categories. This
-            #       will make documentation much simpler.
-            name = name + '_' + category
-
-        if name in attrs:
-            raise ValueError('{0} `{1}` would clobber existing '
-                             '{2} resource attribute'.format(
-                                category, name, resource_name))
-
-        return name
+            attrs[waiter.name] = self._create_waiter(waiter)
 
     def _create_autoload_property(factory_self, name, snake_cased):
         """
@@ -272,35 +203,35 @@ class ResourceFactory(object):
         # it first checks to see if it CAN be loaded (raise if not), then
         # calls the load before returning the value.
         def property_loader(self):
-            if self.meta['data'] is None:
+            if self.meta.data is None:
                 if hasattr(self, 'load'):
                     self.load()
                 else:
                     raise ResourceLoadException(
                         '{0} has no load method'.format(self.__class__.__name__))
 
-            return self.meta['data'].get(name)
+            return self.meta.data.get(name)
 
         property_loader.__name__ = str(snake_cased)
         property_loader.__doc__ = 'TODO'
         return property(property_loader)
 
-    def _create_waiter(factory_self, waiter_model, snake_cased):
+    def _create_waiter(factory_self, waiter_model):
         """
         Creates a new wait method for each resource where both a waiter and
         resource model is defined.
         """
-        waiter = WaiterAction(waiter_model, waiter_resource_name=snake_cased)
+        waiter = WaiterAction(waiter_model,
+                              waiter_resource_name=waiter_model.name)
         def do_waiter(self, *args, **kwargs):
             waiter(self, *args, **kwargs)
 
-        do_waiter.__name__ = str(snake_cased)
+        do_waiter.__name__ = str(waiter_model.name)
         do_waiter.__doc__ = 'TODO'
         return do_waiter
 
     def _create_collection(factory_self, service_name, resource_name,
-                           snake_cased, collection_model,
-                           resource_defs, service_model):
+                           collection_model, resource_defs, service_model):
         """
         Creates a new property on the resource to lazy-load a collection.
         """
@@ -312,40 +243,47 @@ class ResourceFactory(object):
             return cls(collection_model, self, factory_self,
                        resource_defs, service_model)
 
-        get_collection.__name__ = str(snake_cased)
+        get_collection.__name__ = str(collection_model.name)
         get_collection.__doc__ = 'TODO'
         return property(get_collection)
 
-    def _create_reference(factory_self, name, snake_cased, reference,
-                          service_name, resource_name, model, resource_defs,
-                          service_model):
+    def _create_reference(factory_self, name, reference, service_name,
+                          resource_name, model, resource_defs, service_model):
         """
         Creates a new property on the resource to lazy-load a reference.
         """
+        # References are essentially an action with no request
+        # or response, so we can re-use the response handlers to
+        # build up resources from identifiers and data members.
+        handler = ResourceHandler(reference.resource.path, factory_self,
+                                  resource_defs, service_model,
+                                  reference.resource)
+
+        # Are there any identifiers that need access to data members?
+        # This is important when building the resource below since
+        # it requires the data to be loaded.
+        needs_data = any(i.source == 'data' for i in
+                         reference.resource.identifiers)
+
         def get_reference(self):
             # We need to lazy-evaluate the reference to handle circular
             # references between resources. We do this by loading the class
             # when first accessed.
-            # First, though, we need to see if we have the required
-            # identifiers to instantiate the resource reference.
-            identifiers = build_identifiers(
-                reference.resource.identifiers, self, {}, {})
-            resource = None
-            if all_not_none(identifiers.values()):
-                # Identifiers are present, so now we can create the resource
-                # instance using them.
-                resource_type = reference.resource.type
-                cls = factory_self.load_from_definition(
-                    service_name, name, resource_defs.get(resource_type),
-                    resource_defs, service_model)
-                resource = cls(**identifiers)
-            return resource
+            # This is using a *response handler* so we need to make sure
+            # our data is loaded (if possible) and pass that data into
+            # the handler as if it were a response. This allows references
+            # to have their data loaded properly.
+            if needs_data and self.meta.data is None and hasattr(self, 'load'):
+                self.load()
+            return handler(self, {}, self.meta.data)
 
-        get_reference.__name__ = str(snake_cased)
+        get_reference.__name__ = str(reference.name)
         get_reference.__doc__ = 'TODO'
         return property(get_reference)
 
-    def _create_class_partial(factory_self, resource_cls, identifiers=None):
+    def _create_class_partial(factory_self, name, subresource,
+                              service_name, resource_name, model,
+                              resource_defs, service_model):
         """
         Creates a new method which acts as a functools.partial, passing
         along the instance's low-level `client` to the new resource
@@ -354,39 +292,31 @@ class ResourceFactory(object):
         # We need a new method here because we want access to the
         # instance's client.
         def create_resource(self, *args, **kwargs):
-            pargs = []
+            positional_args = []
+
+            # We lazy-load the class to handle circular references.
+            resource_cls = factory_self.load_from_definition(
+                service_name, name, resource_defs.get(name, {}),
+                resource_defs, service_model)
 
             # Assumes that identifiers are in order, which lets you do
             # e.g. ``sqs.Queue('foo').Message('bar')`` to create a new message
             # linked with the ``foo`` queue and which has a ``bar`` receipt
             # handle. If we did kwargs here then future positional arguments
             # would lead to failure.
+            identifiers = subresource.resource.identifiers
             if identifiers is not None:
-                for key, value in identifiers.items():
-                    pargs.append(getattr(self, xform_name(key)))
+                for identifier, value in build_identifiers(identifiers, self):
+                    positional_args.append(value)
 
-            return partial(resource_cls, *pargs,
-                client=self.meta.get('client'))(*args, **kwargs)
+            return partial(resource_cls, *positional_args,
+                client=self.meta.client)(*args, **kwargs)
 
-        # Generate documentation about required and optional params
-        doc = 'Create a new instance of {0}\n\nRequired identifiers:\n'
-
-        for identifier in resource_cls.meta['identifiers']:
-            doc += ':type {0}: string\n'.format(identifier)
-            doc += ':param {0}: {0} identifier\n'.format(identifier)
-
-        doc += '\nOptional params:\n'
-        doc += ':type client: botocore.client\n'
-        doc += ':param client: Low-level Botocore client instance\n'
-
-        doc += '\n:rtype: {0}\n'.format(resource_cls)
-        doc += ':return: A new resource instance'
-
-        create_resource.__name__ = str(resource_cls.__name__)
-        create_resource.__doc__ = doc.format(resource_cls)
+        create_resource.__name__ = str(name)
+        create_resource.__doc__ = 'TODO'
         return create_resource
 
-    def _create_action(factory_self, snake_cased, action_model, resource_defs,
+    def _create_action(factory_self, action_model, resource_defs,
                        service_model, is_load=False):
         """
         Creates a new method which makes a request to the underlying
@@ -405,7 +335,7 @@ class ResourceFactory(object):
             # instance via ``self``.
             def do_action(self, *args, **kwargs):
                 response = action(self, *args, **kwargs)
-                self.meta['data'] = response
+                self.meta.data = response
         else:
             # We need a new method here because we want access to the
             # instance via ``self``.
@@ -416,10 +346,10 @@ class ResourceFactory(object):
                     # Clear cached data. It will be reloaded the next
                     # time that an attribute is accessed.
                     # TODO: Make this configurable in the future?
-                    self.meta['data'] = None
+                    self.meta.data = None
 
                 return response
 
-        do_action.__name__ = str(snake_cased)
+        do_action.__name__ = str(action_model.name)
         do_action.__doc__ = 'TODO'
         return do_action
