@@ -14,10 +14,11 @@
 import os
 
 import botocore.session
+from botocore.client import Config
 
 import boto3
+import boto3.utils
 
-from .exceptions import NoVersionFound
 from .resources.factory import ResourceFactory
 
 
@@ -37,15 +38,18 @@ class Session(object):
     :type botocore_session: botocore.session.Session
     :param botocore_session: Use this Botocore session instead of creating
                              a new default one.
+    :type profile_name: string
+    :param profile_name: The name of a profile to use. If not given, then
+                         the default profile is used.
     """
     def __init__(self, aws_access_key_id=None, aws_secret_access_key=None,
                  aws_session_token=None, region_name=None,
-                 botocore_session=None):
+                 botocore_session=None, profile_name=None):
         if botocore_session is not None:
             self._session = botocore_session
         else:
             # Create a new default session
-            self._session = botocore.session.Session()
+            self._session = botocore.session.get_session()
 
         # Setup custom user-agent string if it isn't already customized
         if self._session.user_agent_name == 'Botocore':
@@ -58,6 +62,9 @@ class Session(object):
             self._session.user_agent_name = 'Boto3'
             self._session.user_agent_version = boto3.__version__
 
+        if profile_name is not None:
+            self._session.set_config_variable('profile', profile_name)
+
         if aws_access_key_id or aws_secret_access_key or aws_session_token:
             self._session.set_credentials(aws_access_key_id,
                 aws_secret_access_key, aws_session_token)
@@ -65,62 +72,29 @@ class Session(object):
         if region_name is not None:
             self._session.set_config_variable('region', region_name)
 
-        self.resource_factory = ResourceFactory()
+        self.resource_factory = ResourceFactory(
+            self._session.get_component('event_emitter'))
         self._setup_loader()
+        self._register_default_handlers()
 
     def __repr__(self):
         return 'Session(region={0})'.format(
             repr(self._session.get_config_variable('region')))
 
+    @property
+    def profile_name(self):
+        """
+        The **read-only** profile name.
+        """
+        return self._session.profile or 'default'
+
     def _setup_loader(self):
         """
         Setup loader paths so that we can load resources.
         """
         self._loader = self._session.get_component('data_loader')
-        self._loader.data_path = ':'.join(
-            [self._loader.data_path,
-             os.path.join(os.path.dirname(__file__), 'data',
-                          'resources')]).strip(':')
-
-    def _get_resource_files(self):
-        """
-        This generator yields paths to resource files in the loader's
-        search paths. Specifically, it looks for files that end with
-        ``.resources.json`` in any of the search paths, but does not
-        recursively search the paths.
-        """
-        for path in self._loader.get_search_paths():
-            if not os.path.isdir(path) or not os.path.exists(path):
-                continue
-
-            items = os.listdir(path)
-            for entry in [i for i in items if i.endswith('.resources.json')]:
-                yield entry
-
-    def _find_latest_version(self, service_name):
-        """
-        Find the latest resource version of a given service if it exists,
-        otherwise raises an exception.
-
-        TODO: Merge this logic upstream into Botocore if possible. Botocore
-              depends on a different directory layout at the moment.
-
-        :rtype: string
-        :return: Version string like 'YYYY-MM-DD'
-        :raises: NoVersionFound
-        """
-        filtered = []
-        for path in self._get_resource_files():
-            if path.startswith(service_name + '-'):
-                filtered.append(path)
-
-        try:
-            # ['s3-2006-03-01.resources.json', ...] => '2006-03-01'
-            # Hard coded offsets below pull out just the date string
-            start = len(service_name)
-            return max([i[start + 1:start + 11] for i in filtered])
-        except ValueError:
-            raise NoVersionFound(service_name)
+        self._loader.search_paths.append(
+             os.path.join(os.path.dirname(__file__), 'data'))
 
     def get_available_services(self):
         """
@@ -140,22 +114,12 @@ class Session(object):
         :rtype: list
         :return: List of service names
         """
-        service_names = set()
-
-        for path in self._get_resource_files():
-            # TODO: Glacier is not yet supported by Botocore.
-            if 'glacier' in path:
-                continue
-
-            # 'foo-bar-2006-03-01' => 'foo-bar'
-            service_names.add('-'.join(path.split('-')[:-3]))
-
-        return list(service_names)
+        return self._loader.list_available_services(type_name='resources-1')
 
     def client(self, service_name, region_name=None, api_version=None,
                use_ssl=True, verify=None, endpoint_url=None,
                aws_access_key_id=None, aws_secret_access_key=None,
-               aws_session_token=None):
+               aws_session_token=None, config=None):
         """
         Create a low-level service client by name.
 
@@ -211,14 +175,21 @@ class Session(object):
         :param aws_session_token: The session token to use when creating
             the client.  Same semantics as aws_access_key_id above.
 
+        :type config: botocore.client.Config
+        :param config: Advanced client configuration options. If region_name
+            is specified in the client config, its value will take precedence
+            over environment variables and configuration values, but not over
+            a region_name value passed explicitly to the method.
+
         :return: Service client instance
+
         """
         return self._session.create_client(
             service_name, region_name=region_name, api_version=api_version,
             use_ssl=use_ssl, verify=verify, endpoint_url=endpoint_url,
             aws_access_key_id=aws_access_key_id,
             aws_secret_access_key=aws_secret_access_key,
-            aws_session_token=aws_session_token)
+            aws_session_token=aws_session_token, config=config)
 
     def resource(self, service_name, region_name=None, api_version=None,
                use_ssl=True, verify=None, endpoint_url=None,
@@ -281,21 +252,44 @@ class Session(object):
 
         :return: Subclass of :py:class:`~boto3.resources.base.ServiceResource`
         """
+        if api_version is None:
+            api_version = self._loader.determine_latest_version(
+                service_name, 'resources-1')
+        resource_model = self._loader.load_service_model(
+            service_name, 'resources-1', api_version)
         # Creating a new resource instance requires the low-level client
         # and service model, the resource version and resource JSON data.
         # We pass these to the factory and get back a class, which is
         # instantiated on top of the low-level client.
+        config = Config(user_agent_extra='Resource')
         client = self.client(
             service_name, region_name=region_name, api_version=api_version,
             use_ssl=use_ssl, verify=verify, endpoint_url=endpoint_url,
             aws_access_key_id=aws_access_key_id,
             aws_secret_access_key=aws_secret_access_key,
-            aws_session_token=aws_session_token)
-        service_model = self._session.get_service_model(service_name)
-        version = self._find_latest_version(service_name)
-        model = self._loader.load_data(
-            '{0}-{1}.resources'.format(service_name, version))
+            aws_session_token=aws_session_token, config=config)
+        service_model = client.meta.service_model
         cls = self.resource_factory.load_from_definition(
-            service_name, service_name, model['service'], model['resources'],
-            service_model)
+            service_name, service_name, resource_model['service'],
+            resource_model['resources'], service_model)
         return cls(client=client)
+
+    def _register_default_handlers(self):
+        self._session.register(
+            'creating-client-class.s3',
+            boto3.utils.lazy_call(
+                'boto3.s3.inject.inject_s3_transfer_methods'))
+        self._session.register(
+            'creating-resource-class.s3.Bucket',
+            boto3.utils.lazy_call(
+                'boto3.s3.inject.inject_bucket_load'))
+        self._session.register(
+            'creating-resource-class.dynamodb',
+            boto3.utils.lazy_call(
+                'boto3.dynamodb.transform.register_high_level_interface'),
+            unique_id='high-level-dynamodb')
+        self._session.register(
+            'creating-resource-class.dynamodb.Table',
+            boto3.utils.lazy_call(
+                'boto3.dynamodb.table.register_table_methods'),
+            unique_id='high-level-dynamodb-table')
